#ifndef __CORE_SW2_NODE_H_
#define __CORE_SW2_NODE_H_

#define CG_NUM		4

#define SW_PORT_CONTROL         (0x6fff00UL | PAGE_OFFSET)
#define SW2_PCI_MEM_UNPRE	0x0UL

#ifdef CONFIG_SW6A
#ifndef CONFIG_SICH_VIDEO
#define SW2_PCI_MEM_PRE         0x3c0000000UL
#define SW2_PCI_MEM_PRI_SIZE    0x40000000UL
#else
#define SW2_PCI_MEM_PRE         0x200000000UL
#define SW2_PCI_MEM_PRI_SIZE    0x200000000UL
#endif

#define SW2_PCI_IO_PIU0	0x400000000UL
#define SW2_PCI_IO_PIU1	0x500000000UL
#define SW2_PCI_IO_EP		0x600000000UL
#define SW2_PCI_EP_CONFIG	0xa00000000UL
#define SW2_PCI_MEM		0xe0000000UL
#define SW2_PCI_MEM_SIZE        0x20000000UL
#define SW2_PCI_IO_SIZE         0xf00000UL
#define SW2_PCI_IO_RC		0x800000000UL

#ifdef CONFIG_SW_4LEVEL_PGTABLE
/* Define each pcie base address */
#ifdef CONFIG_SW_QEMU
#define SW2_LEGACY_IO		0x804600000000UL
#else
#define SW2_LEGACY_IO		0x807600000000UL
#endif
#define SW2_IO_PCIE(n)          ((1UL << 47) + 0x4000000000UL + (0x3UL << 36) * (n))
#else /*!CONFIG_SW_4LEVEL_PGTABLE*/
#define SW2_LEGACY_IO           0xf600000000
#define SW2_IO_PCIE(n)          (0xc000000000UL + (0x3UL << 36) * (n))
#endif

#define SW2_PCIE_IO_BASE(n)		(SW2_IO_PCIE(n) + SW2_PCI_IO_EP)
#define SW2_PCIE_MEM_BASE(n)		(SW2_IO_PCIE(n) + SW2_PCI_MEM)
#define SW2_PCIE_MEM_PRI_BASE(n)	(SW2_IO_PCIE(n) + SW2_PCI_MEM_PRE)
#endif /* CONFIG_SW_6 */



#define MSI_FIRST_VECTOR 15

/* PIU internal register define */
#define ERR_REPORT		0x0
#define CXCPL_DEBUG_INFO	0x5
#define MSI_ADDR		0x9

/* Define each PCIE start scanning bus number */
#ifdef CONFIG_SW_QEMU
#define SW2_PCI0_BUS	0x0  /* QEMU will start from bus 0 */
#else
#define SW2_PCI0_BUS	0x1
#endif
#define SW2_PCI1_BUS	0x20


#define SW2_PCI_DEBUG 0
#if SW2_PCI_DEBUG
#define PCIINFO(fmt, args...)	printk(fmt, ##args)
#else
#define PCIINFO(fmt, args...)
#endif

#ifndef __EXTERN_INLINE
#define __EXTERN_INLINE extern inline
#endif

static inline unsigned long get_RC_config(int index)
{
	return SW2_IO_PCIE(index) + PAGE_OFFSET + SW2_PCI_IO_RC;
}

static inline unsigned long get_EP_PCI_config(int index)
{
	return SW2_IO_PCIE(index) + PAGE_OFFSET + SW2_PCI_EP_CONFIG + 0x100000000;
}

static inline unsigned long get_PIU_PCI_IOR0(int index)
{
	return SW2_IO_PCIE(index) + PAGE_OFFSET + SW2_PCI_IO_PIU0;
}

static inline unsigned long get_PIU_PCI_IOR1(int index)
{
	return SW2_IO_PCIE(index) + PAGE_OFFSET + SW2_PCI_IO_PIU1;
}

static inline unsigned long get_PCI_config(int index)
{
	return SW2_IO_PCIE(index) + PAGE_OFFSET + SW2_PCI_EP_CONFIG;
}

extern inline void __iomem *sw2_ioportmap(unsigned long addr)
{
	return sw_mv.mv_ioportmap(addr);
}

static inline void __iomem *sw2_iounmap(volatile void __iomem *addr)
{
	unsigned long tmp = (unsigned long) addr;

	return (void __iomem *)(tmp - PAGE_OFFSET);
}

static inline void __iomem *sw2_ioremap(unsigned long addr, unsigned long size)
{
	unsigned long tmp = addr + PAGE_OFFSET;

	return (void __iomem *)(tmp);
}

static inline int sw2_is_ioaddr(unsigned long addr)
{
	return addr >= (PAGE_OFFSET + SW2_IO_PCIE(0));
}

static inline int sw2_is_mmio(const volatile void __iomem *xaddr)
{
	unsigned long addr = (unsigned long) xaddr;

	return (addr & 0x100000000UL) == 0;
}


#undef __IO_PREFIX
#define __IO_PREFIX             sw2

#define sw2_trivial_rw_bw   1
#define sw2_trivial_rw_lq   1
#define sw2_trivial_io_bw   1
#define sw2_trivial_io_lq   1

#include <asm/io_trivial.h>

static inline void
sw2_read_rc_conf(int pci_index, unsigned int conf_offset, volatile unsigned int *data)
{
	unsigned long addr;

	addr = get_RC_config(pci_index) | conf_offset;
	PCIINFO("RC config read, address = %#lx,", addr);
	*data = *(unsigned int *)addr;
	mb();
	PCIINFO("value = %#x.\n", *data);
}


static inline void
sw2_write_rc_conf(int pci_index, unsigned int conf_offset, unsigned int data)
{
	unsigned long addr;

	addr = get_RC_config(pci_index) | conf_offset;
	PCIINFO("RC config write, address = %#lx, data = %#x.\n", addr, data);
	*(unsigned int *)addr = data;
	mb();
}

static inline void
sw2_write_piu_ior0(int pci_index, unsigned int reg, unsigned long data)
{
	unsigned long addr;

	addr = get_PIU_PCI_IOR0(pci_index) + (reg << 7);
	PCIINFO("PIU0 write, reg = %#x, address = %#lx, data = %#lx.\n", reg, addr, data);
	*(unsigned long __iomem *)addr = data;
	mb();
}

static inline void
sw2_read_piu_ior1(int pci_index, unsigned int reg, volatile unsigned long *value)
{
	unsigned long addr;

	addr = get_PIU_PCI_IOR1(pci_index) + (reg << 7);
	PCIINFO("PIU1 read, reg = %#x, address = %#lx\n", reg, addr);
	*value = *(unsigned long __iomem *)addr;
	mb();
	PCIINFO("value = %#lx.\n", *value);
}
static inline void
sw2_write_piu_ior1(int pci_index, unsigned int reg, unsigned long data)
{
	unsigned long addr;

	addr = get_PIU_PCI_IOR1(pci_index) + (reg << 7);
	PCIINFO("PIU1 write, reg = %#x, address = %#lx, data = %#lx.\n", reg, addr, data);
	*(unsigned long __iomem *)addr = data;
	mb();
}

static inline void
sw2_read_piu_ior0(int pci_index, unsigned int reg, volatile unsigned long *value)
{
	unsigned long addr;

	addr = get_PIU_PCI_IOR0(pci_index) + (reg << 7);
	PCIINFO("PIU0 read, reg = %#x, address = %#lx,", reg, addr);
	*value = *(unsigned long __iomem *)addr;
	mb();
	PCIINFO("value = %#lx.\n", *value);
}

#endif /* __CORE_SW2_NODE_H_ */
