#ifndef _CROSSM_H_
#define  _CROSSM_H_

#ifdef __KERNEL__
#include <linux/delay.h>

/* CROSS_PADDR : PA[45] = 1 */
#define CROSS_PADDR_BASE	(0x1UL << 45)

#define CROSS_SHIFT		9
#define CROSS_GRAIN		(0x1UL << CROSS_SHIFT)
#define CROSS_MASK		(~(CROSS_GRAIN - 1))

static inline unsigned long cross_paddr_to_cg_paddr(unsigned long cross_paddr)
{
	unsigned long temp_paddr = cross_paddr & (CROSS_PADDR_BASE - 1);
	unsigned long cg_num = (temp_paddr >> CROSS_SHIFT) & 0x3;
	/* 交叉块地址 */
	unsigned long cross_block_addr = (temp_paddr >> (CROSS_SHIFT + 2)) << CROSS_SHIFT;
	/* 交叉块内偏移地址 */
	unsigned long cross_block_offset = temp_paddr & (~CROSS_MASK);

	return cross_block_addr  | cross_block_offset | (cg_num << SW2_CG_MEM_SHIFT);
}

static inline unsigned long cg_paddr_to_cross_paddr(unsigned long cg_paddr)
{
	unsigned long cg_num = (cg_paddr >> SW2_CG_MEM_SHIFT) & 0x3;
	unsigned long cross_block_addr = (cg_paddr >> CROSS_SHIFT) << (CROSS_SHIFT + 2);
	unsigned long cross_block_offset = cg_paddr & (~CROSS_MASK);

	return cross_block_addr | (cg_num << CROSS_SHIFT) | cross_block_offset | CROSS_PADDR_BASE;
}

/* 各个核组上的连续区bottom地址 */
#define cont_bottom_cg_paddr(cont_cpm, cg_num) \
(cont_cpm | ((long)cg_num * (1UL << SW2_CG_MEM_SHIFT)))

#define for_each_busy_crossm(crossm) \
	list_for_each_entry(crossm, &crossm_list_busy, crossm_list)
#define for_each_idle_crossm(crossm) \
	list_for_each_entry(crossm, &crossm_list_idle, crossm_list)

#define cg_online MCU_READ_REG(MC_ONLINE)

#define for_each_online_cg(i) \
	for (i = 0; i < NR_SCORE_ARRAY_NUM; i++) \
		if ((cg_online >> i) & 0x1)

#define cg_online_num()		\
({					\
	int i, cg_num = 0;		\
	for_each_online_cg(i)		\
		cg_num++;		\
	cg_num;				\
})

struct crossm *find_busy_crossm_by_jobid(int jobid);
int do_sjob_free_crossm(struct sjob *sjob);
int do_sjob_alloc_crossm(struct file *filp, struct sjob *sjob, struct sjob_alloc_crossm *alloc_crossm);

/*
static inline void wait_flush_scache_all(void)
{
    // TODO
}

static inline void flush_scache_all_end(void)
{
    // TODO
}
*/

#endif

#endif  /* ----- #ifndef _CROSSM_H_  ----- */
